/*
** Copyright (C) 2003-2006 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "exception.h"
#include "libraries.h"
#include <glib.h>
#include "sqlite_reader.h"
#include <sqlite3.h>
#include "shell.h"
#include "gwrappers.h"
#include "utilities.h"
#include "bible.h"
#include "books.h"


ustring database ();
void wait_till_database_present_and_initialized ();
bool we_ought_to_run ();
void main_loop ();
void log (const ustring & message);
gchar * dcop ();
void getreference ();
void setreference ();
void dcop_arguments (GwSpawn& spawn);
void check_exit (bool spawned, int exitstatus);
void getmodules ();
void reloadmodules ();
void set_connected (bool connect);
void search ();


#define STAGE_ZERO 0
#define STAGE_COMMUNICATE 1
#define STAGE_WAIT_RETRY 1000
#define STAGE_RETRY 1100
#define BIBLETIME "bibletime"


int stage = STAGE_ZERO;
ustring previous_message;
ustring bibledit_dcop_name;
bool connected = false;
ustring reply;


ustring database ()
// Returns the filename of the database for communication with bibledit.
{
  ustring filename;
  gchar * name = g_build_filename (g_get_home_dir(), ".bibledit_temp", "bibletime.sql", NULL);
  filename = name;
  g_free (name);
  return filename;
}


void wait_till_database_present_and_initialized ()
{
  // Wait till the database file is there. 
  while (!g_file_test (database ().c_str(), G_FILE_TEST_IS_REGULAR)) {
    g_usleep (100000);
  }
  g_usleep (100000);
  // Wait further till that database has been initialized.
  while (!we_ought_to_run ()) {
    g_usleep (100000);
  }
}


bool we_ought_to_run ()
// Returns true if we ought to run.
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql = g_strdup_printf ("select run from runtime;");
  SqliteReader reader (0);
  rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
  bool run = !reader.ustring0.empty();
  g_free (sql);
  sqlite3_close (db);
  return run;
}


void main_loop ()
{
  switch (stage) {
    case STAGE_ZERO:
    {
      // Check if BibleTime is running.
      if (!program_is_running_basic (BIBLETIME)) {
        log ("BibleTime is not running");
        stage = STAGE_WAIT_RETRY;
        break;
      }
      // Check if dcop protocol is working, and get bibletime's name.
      bibledit_dcop_name.clear();
      GwSpawn spawn (dcop());
      spawn.read ();
      spawn.run ();
      for (unsigned int i = 0; i < spawn.standardout.size(); i++) {
        spawn.standardout[i] = trim (spawn.standardout[i]);
        if (spawn.standardout[i].find (BIBLETIME) != string::npos) {
          bibledit_dcop_name = spawn.standardout[i];
        }
      }
      // Was bibledit's dcop name found?
      if (bibledit_dcop_name.empty()) {
        log ("No DCOP connection possible");
        stage = STAGE_WAIT_RETRY;
        break;
      }
      // Ok, connection is possible.
      log ("Connected");
      set_connected (true);
      stage = STAGE_COMMUNICATE;

    }
    case STAGE_COMMUNICATE:
    {
      getreference ();
      setreference ();
      getmodules ();
      search ();
      reloadmodules ();
      break;
    }
    case STAGE_WAIT_RETRY:
    {
      stage++;
      break;
    }
    case STAGE_RETRY:
    {
      stage = STAGE_ZERO;
      break;
    }
    default:
    {
      stage++;
    }
  }
  // Delay a bit.
  g_usleep (300000);
}


void log (const ustring & message)
{
  if (message != previous_message) {
    gw_message (message);
    previous_message = message;
  }
}


gchar * dcop ()
{
  return "dcop";
}


void getreference ()
// Get the reference from BibleTime. Store it in the database.
{
  // Only communicate when BibleTime is connected.
  if (connected) {
    // Store BibleTime's reply.
    ustring result;
    // Execute command to elicit a reply.
    GwSpawn spawn (dcop ());
    dcop_arguments (spawn);
    spawn.arg ("getCurrentReference");
    spawn.read ();
    spawn.run ();
    if (!spawn.standardout.empty())
      result = spawn.standardout[0];
    check_exit (true, spawn.exitstatus); 
    // Only proceed if we have a different value than before.
    if (result != reply) {
      reply = result;
      if (!result.empty()) log ("Received " + result);
      // Clean up the reply and see whether we have either a Bible or a Commentary.
      result = trim (result);
      #define BIBLETIME_BIBLE "[BIBLE]"
      #define BIBLETIME_COMMENTARY "[COMMENTARY]"
      size_t offset;
      offset = result.find (BIBLETIME_BIBLE);
      if (offset != string::npos) {
        offset += strlen (BIBLETIME_BIBLE);
      } else {
        offset = result.find (BIBLETIME_COMMENTARY);
        if (offset != string::npos) {
          offset += strlen (BIBLETIME_COMMENTARY);
        }
      }
      #undef BIBLETIME_BIBLE
      #undef BIBLETIME_COMMENTARY
      // In case of Bible/commentary, transform the reference to our format.
      // Store the result in the database, if we got something. Else erase it.
      if (offset != string::npos) {
        result.erase (0, offset);
        result = trim (result);
        sqlite3 *db;
        int rc;
        rc = sqlite3_open (database ().c_str (), &db);
        sqlite3_busy_timeout (db, 1000);
        char * sql;
        sql = g_strdup_printf ("delete from getref;");
        rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
        g_free (sql);
        unsigned int book, chapter;
        ustring verse;
        if (reference_discover (0, 0, "", result, book, chapter, verse)) {
          sql = g_strdup_printf ("insert into getref values (%d, %d, '%s');", book, chapter, verse.c_str());
        }
        rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
        g_free (sql);
        sqlite3_close (db);
      }
    }
  } 
}


void setreference ()
// If the database has a reference, send it to BibleTime.
{
  // Look in the database whether there is any reference to be sent to 
  // BibleTime. If there is nothing, bail out.
  SqliteReader reader (0);
  bool noreference;
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql = g_strdup_printf ("select book, chapter, verse from setref;");
    rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
    noreference = reader.ustring0.empty();
    g_free (sql);
    sqlite3_close (db);
  }
  if (noreference) return;
  unsigned int book = convert_to_int (reader.ustring0[0]);
  ustring chapter = reader.ustring1[0];
  ustring verse = reader.ustring2[0];

  // Remove the command from the database.
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql = g_strdup_printf ("delete from setref;");
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    g_free (sql);
    sqlite3_close (db);
  }
  
  // Bail out if disconnected.
  if (!connected) return;
  
  // Send the reference to BibleTime.
  // Handle case that project is closed, so the book is 0. Solves a crash bug.
  if (!book) return;
  ustring osisreference = books_id_to_osis (book) + "." + chapter + "." + verse;
  
  GwSpawn spawn (dcop ());
  dcop_arguments (spawn);
  spawn.arg ("syncAllVerseBasedModules");
  spawn.arg (osisreference);
  spawn.run ();
  check_exit (spawn.result, spawn.exitstatus);
  log ("Sent: " + osisreference);
}


void dcop_arguments (GwSpawn& spawn)
{
  spawn.arg (bibledit_dcop_name);
  spawn.arg ("BibleTimeInterface");
}


void check_exit (bool spawned, int exitstatus)
{
  if (!spawned || (exitstatus != 0)) {
    set_connected (false);
    stage = STAGE_ZERO;
    log ("Error while communicating");
  }
}


void getmodules ()
{
  // Bail out if disconnected.
  if (!connected) return;

  // Database variables.
  sqlite3 *db;
  int rc;
  char * sql;
  
  // Look in the database whether the modules are to be fetched. 
  // If not, bail out.
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  sql = g_strdup_printf ("select flag from getmodules;");
  SqliteReader reader (0);
  rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
  bool nothing = reader.ustring0.empty();
  g_free (sql);
  sqlite3_close (db);
  if (nothing) return;

  // Clear relevant flag, the bibles and the commentaries in database.
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  sql = g_strdup_printf ("delete from getmodules;");
  rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("delete from bibles;");
  rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("delete from commentaries;");
  rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
  g_free (sql);
  
  // Get the Bibles from BibleTime.
  {
    ustring message = "Bibles:";
    GwSpawn spawn (dcop ());
    dcop_arguments (spawn);
    spawn.arg ("getModulesOfType");
    spawn.arg ("BIBLES");
    spawn.read ();
    spawn.run ();
    for (unsigned int i = 0; i < spawn.standardout.size(); i++) {
      ustring result = spawn.standardout[i];
      result = trim (result);
      if (!result.empty()) {
        sql = g_strdup_printf ("insert into bibles values ('%s');", result.c_str());
        rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
        g_free (sql);
        if (i) message.append (",");
        message.append (" " + result);
      }
    }
    log (message);
  }

  // Get the Commentaries from BibleTime.
  {
    ustring message = "Commentaries:";
    GwSpawn spawn (dcop ());
    dcop_arguments (spawn);
    spawn.arg ("getModulesOfType");
    spawn.arg ("COMMENTARIES");
    spawn.read ();
    spawn.run ();
    for (unsigned int i = 0; i < spawn.standardout.size(); i++) {
      ustring result = spawn.standardout[i];
      result = trim (result);
      if (!result.empty()) {
        sql = g_strdup_printf ("insert into commentaries values ('%s');", result.c_str());
        rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
        g_free (sql);
        if (i) message.append (",");
        message.append (" " + result);
      }
    }
    log (message);
  }
  
  // Close database.
  sqlite3_close (db);
}


void reloadmodules ()
// Sends a signal to BibleTime to reload all the modules.
{
  // Bail out if disconnected.
  if (!connected) return;

  // Look in the database whether the modules are to be reloaded. 
  // If not, bail out.
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql = g_strdup_printf ("select flag from reloadmodules;");
    SqliteReader reader (0);
    rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
    bool nothing = reader.ustring0.empty();
    g_free (sql);
    sqlite3_close (db);
    if (nothing) return;
  }

  // Remove the command from the database.
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql = g_strdup_printf ("delete from reloadmodules;");
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    g_free (sql);
    sqlite3_close (db);
  }

  // Get the Bibles from BibleTime.
  log ("Reloading modules");
  GwSpawn spawn (dcop ());
  dcop_arguments (spawn);
  spawn.arg ("reloadModules");
  spawn.run ();
}


void set_connected (bool connect)
// Writes to the database whether BibleTime is connected.
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql;
  sql = g_strdup_printf ("delete from connected;");
  rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
  g_free (sql);
  if (connect) {
    sql = g_strdup_printf ("insert into connected values (1);");
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    g_free (sql);
  }
  sqlite3_close (db);
  connected = connect;
}


void search ()
{
  // Bail out if disconnected.
  if (!connected) return;

  // Look in the database whether a search has to be made. 
  // If not, bail out.
  SqliteReader reader (0);
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql = g_strdup_printf ("select run, module, text, selector from searchcommand;");
    rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
    bool nothing = reader.ustring0.empty();
    g_free (sql);
    sqlite3_close (db);
    if (nothing) return;
  }

  // Delete the search commands from the database.
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql = g_strdup_printf ("delete from searchcommand;");
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    g_free (sql);
    sqlite3_close (db);
  }
  
  // Assemble the search variables.
  ustring module = reader.ustring1[0];
  ustring text = reader.ustring2[0];
  unsigned int selector = convert_to_int (reader.ustring3[0]);
  
  // Create the command to send to BibleTime in order to search.
  vector <ustring> search_command;
  switch (selector) {
    case 0 :
      search_command.push_back ("searchInDefaultBible");
      break;
    case 1 :
      search_command.push_back ("searchInOpenModules");
      break;
    case 2 :
      search_command.push_back ("searchInModule");
      search_command.push_back (module);
      break;
  }

  // Add searchtext to the command.  
  // Escape certain characters in the text, because it goes through the shell.
  gchar * escaped_text = g_strescape (text.c_str(), "");
  ustring searchtext = escaped_text;
  g_free (escaped_text);
  search_command.push_back (searchtext);
  
  // Get the search results from BibleTime.
  vector<ustring> search_results;  
  GwSpawn spawn (dcop ());
  dcop_arguments (spawn);
  for (unsigned int i = 0; i < search_command.size(); i++) {
    spawn.arg (search_command[i]);
  }
  spawn.read ();
  spawn.run ();
  for (unsigned int i = 0; i < spawn.standardout.size(); i++) {
    ustring result = spawn.standardout[i];
    result = trim (result);
    if (!result.empty()) {
      search_results.push_back (result);
    }
  }

  // Write these results to the database.
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql;
    sql = g_strdup_printf ("PRAGMA synchronous=OFF;");
    rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
    g_free (sql);
    sql = g_strdup_printf ("delete from searchresults;");
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    g_free (sql);
    for (unsigned int i = 0; i < search_results.size(); i++) {
      sql = g_strdup_printf ("insert into searchresults values ('%s');", double_apostrophy (search_results[i]).c_str());
      rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
      g_free (sql);      
    }
    sql = g_strdup_printf ("insert into searchdone values (1);");
    rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
    g_free (sql);
    sqlite3_close (db);
  }
 
  // Log  
  log ("Search for " + text + ": " + convert_to_string (search_results.size()) + " results");
}


int main (int argc, char *argv[])
{
  // Wait till things are ready to go.
  wait_till_database_present_and_initialized ();
  
  // While the database indicates that we ought to run, we keep going.
  while (we_ought_to_run ()) {
    main_loop ();    
  }
  
  // The job's been done.
  unlink (database().c_str());
  return 0;
}
