/*
** Copyright (C) 2003-2006 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "exception.h"
#include "libraries.h"
#include <glib.h>
#include "sqlite_reader.h"
#include <sqlite3.h>
#include "shell.h"
#include "gwrappers.h"
#include "utilities.h"
#include "bible.h"
#include "books.h"
#include "subversion.h"
#include "types.h"
#include "projectutils.h"
#include "projectconfig.h"


int main (int argc, char *argv[]);
void wait_till_database_present_and_initialized ();
bool we_ought_to_run ();
bool we_ought_to_pause ();
void main_loop ();
void update_frequency_set (unsigned int frequency);
static bool projects_update_timeout (gpointer data);


int main (int argc, char *argv[])
{
  // Wait till things are ready to go.
  wait_till_database_present_and_initialized ();
  
  // While the database indicates that we ought to run, we keep going.
  while (we_ought_to_run ()) {

    // If pause, handle that.
    if (we_ought_to_pause ()) {
      g_usleep (1000000);
    } else {
      main_loop ();    
    }      
  
  }
  
  // The job's been done.
  unlink (subversion_command_database().c_str());
  gw_message ("Terminated normally");
  return 0;
}


void wait_till_database_present_and_initialized ()
{
  // Wait till the database file is there. 
  while (!g_file_test (subversion_command_database ().c_str(), G_FILE_TEST_IS_REGULAR)) {
    g_usleep (100000);
  }
  // Wait a reasonably long time before subversion activity it starting.
  // This is because Bibledit itself is starting, and if subversion activity
  // would be starting at the same time too, that would pospone further the
  // moment that Bibledit gets responsive.
  g_usleep (5000000);
  // Wait further till that database has been initialized.
  while (!we_ought_to_run ()) {
    g_usleep (100000);
  }
}


bool we_ought_to_run ()
// Returns true if we ought to run.
{
  // It was discovered that at times the busy timeout was exceeded.
  // This caused that it returned "run" as false, and that caused
  // bibledit-subversion to terminate unexpectedly.
  // The exception handler solves it by setting "run" to true in such cases.
  bool run = false;
  sqlite3 *db;
  int rc;
  try
  {
    rc = sqlite3_open (subversion_command_database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql = g_strdup_printf ("select run from flags;");
    SqliteReader reader (0);
    rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
    if (rc) throw runtime_error (sqlite3_errmsg(db));
    if (!reader.ustring0.empty())
      run = convert_to_bool (reader.ustring0[0]);
    g_free (sql);
  }
  catch (exception & ex)
  {
    gw_critical (ex.what ());
    run = true;
  }
  sqlite3_close (db);
  return run;
}


bool we_ought_to_pause ()
// Returns true if we ought to pause.
{
  bool pause = false;
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (subversion_command_database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql = g_strdup_printf ("select pause from flags;");
  SqliteReader reader (0);
  rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
  if (!reader.ustring0.empty())
    pause = convert_to_bool (reader.ustring0[0]);
  g_free (sql);
  sqlite3_close (db);
  return pause;
}


void main_loop ()
/*
This main loop is called repeatedly. Each time it looks whether a task
is available for execution. If there is one, it will execute this task.
It records whether the task was succesful.
*/
{
  // Variables.
  bool task_was_done = false;
  unsigned int task;
  ustring project;
  unsigned int book, chapter;

  // Get the next available task.
  if (subversion_execute_get_next_task (task, project, book, chapter)) {
    
    // Okay, there is something to be done. 
    // Flag "done" as true, so that only in case of an error, it is set false.
    task_was_done = true;

    // Execute the right task.    
    switch ((SubversionTaskType) task) {
      case sttInitializeProject: 
      {
        task_was_done = subversion_execute_initialize_project_local (project);
        break;
      }
      case sttCommitProject:
      {
        task_was_done = subversion_execute_commit_project (project);
        break;
      }
      case sttStoreChapter: 
      {
        task_was_done = subversion_execute_store_chapter (project, book, chapter);
        break;
      }
      case sttUpdateProject:
      {
        task_was_done = subversion_execute_update_project (project);
        break;
      }
      case sttReopenProject:
      { 
        subversion_store_re_open_project (project);
        break;
      }
      case sttOpenChapter:
      { 
        subversion_open_chapter_for_editing (project, book, chapter);
        break;
      }
      case sttUpdateFrequency:
      { 
        // This one is different because the frequency is stored in "book".
        update_frequency_set (book);
        break;
      }
    }
    
    // Erase the task or record a failure.
    if (task_was_done) {
      subversion_execute_erase_task (task, project, book, chapter);
    } else {
      subversion_execute_record_failure (task, project, book, chapter);
      // There are cases that one task hangs, and nearly blocks the program.
      // Introduce a delay to solve this.
      g_usleep (1000000);
    }
  }
  
  // If the task was done, delay shortly so it can quickly do the next one.
  // Else it delays longer to save system resources.
  if (task_was_done) g_usleep (10000);
  else g_usleep (1000000);
}


void update_frequency_set (unsigned int frequency)
// Sets the project update frequency.
{
  // Storate for the event source.
  static int event_source_id = 0;
  // Destroy any previous event source.
  if (event_source_id) {
    GSource *source = g_main_context_find_source_by_id (NULL, event_source_id);
    if (source) 
      g_source_destroy (source);
  }
  // Set the update frequency.  
  event_source_id = g_timeout_add_full (G_PRIORITY_DEFAULT, 1000 * frequency, GSourceFunc (projects_update_timeout), NULL, NULL);
}


bool projects_update_timeout (gpointer data)
// This updates all projects with remote repositories on a regular basis.
{
  vector <ustring> projects = projects_get_all ();
  for (unsigned int i = 0; i < projects.size(); i++) {
    ProjectConfiguration projectconfig (projects[i]);
    if (projectconfig.repository_remote ()) {
      subversion_schedule (sttUpdateProject, projects[i], 0, 0);
    }
  }
  return true;
}
