/*
** Copyright (C) 2003-2006 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include <glib.h>
#include "bibletime.h"
#include "bible.h"
#include "shell.h"
#include "books.h"
#include "session.h"
#include "directories.h"
#include "sqlite_reader.h"


#define STAGE_ZERO 0
#define STAGE_COMMUNICATE 1
#define STAGE_WAIT_RETRY 1000
#define STAGE_RETRY 1200


#define BIBLETIME "bibletime"


BibleTime::BibleTime (bool dummy)
/*
This organizes the communication with BibleTime. It uses KDE's DCOP protocol.
It centralizes all communication with BibleTime so that all happens synchronized,
and two different commands, occurring at the same time, will not disturb each
other.
*/
{
  initialize_database ();
}


BibleTime::~BibleTime ()
{
  finalize_database ();
}


bool BibleTime::connected ()
// Whether BibleTime is connected.
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql = g_strdup_printf ("select flag from connected;");
  SqliteReader reader (0);
  rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
  bool connect = !reader.ustring0.empty();
  g_free (sql);
  sqlite3_close (db);
  return connect;
}


void BibleTime::sendreference (const Reference& reference)
// Stores the reference to be sent to BibleTime in the database. 
// The scheduler will take care of that shortly after.
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql = g_strdup_printf ("delete from setref;");
  rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("insert into setref values (%d, %d, '%s');", reference.book, reference.chapter, reference.verse.c_str());
  rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
  g_free (sql);
  sqlite3_close (db);
}


bool BibleTime::getreference (Reference& reference)
// Read the current reference from the database.
// The database is updated by bibledit-bibletime.
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql = g_strdup_printf ("select book, chapter, verse from getref;");
  SqliteReader reader (0);
  rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
  bool gotreference = !reader.ustring0.empty();
  if (gotreference) {
    reference.book = convert_to_int (reader.ustring0[0]);
    reference.chapter = convert_to_int (reader.ustring1[0]);
    reference.verse = reader.ustring2[0];
  }
  g_free (sql);
  sqlite3_close (db);
  return gotreference;
}


vector<ustring> BibleTime::getbibles ()
{
  // Read the Bibles from the database.
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql = g_strdup_printf ("select name from bibles;");
  SqliteReader reader (0);
  rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
  g_free (sql);
  sqlite3_close (db);

  // Signal the helper to again get the modules.
  getmodules ();

  // Return the Bibles.
  return reader.ustring0;
}


vector<ustring> BibleTime::getcommentaries ()
{
  // Read the commentaries from the database.
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql = g_strdup_printf ("select name from commentaries;");
  SqliteReader reader (0);
  rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
  g_free (sql);
  sqlite3_close (db);

  // Signal the helper to again get the modules.
  getmodules ();

  // Return the commentaries.
  return reader.ustring0;
}


vector<ustring> BibleTime::search_in_default_bible (const ustring& searchtext)
{
  return search ("", searchtext, 0);;
}


vector<ustring> BibleTime::search_in_open_modules (const ustring& searchtext)
{
  return search ("", searchtext, 1);
}


vector<ustring> BibleTime::search_in_module (const ustring& modulename, const ustring& searchtext)
{
  return search (modulename, searchtext, 2);
}


vector<ustring> BibleTime::search (const ustring& modulename, const ustring& searchtext, int selector)
{
  // Write search commands to the database.
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql;
    sql = g_strdup_printf ("insert into searchcommand values (1, '%s', '%s', %d);",
                           double_apostrophy (modulename).c_str(), 
                           double_apostrophy (searchtext).c_str(), 
                           selector);
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    g_free (sql);
    sqlite3_close (db);
  }

  // Wait till the bibledit-bibletime has done the search.
  bool searching = true;
  while (searching) {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql;
    sql = g_strdup_printf ("select flag from searchdone;");
    SqliteReader reader (0);
    rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
    g_free (sql);
    sqlite3_close (db);
    searching = reader.ustring0.empty();
    g_usleep (100000);    
  }
  
  // Search done: erase that flag.
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql;
    sql = g_strdup_printf ("delete from searchdone;");
    rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
    g_free (sql);
    sqlite3_close (db);
  }
  
  // Obtain the search results.
  {
    sqlite3 *db;
    int rc;
    rc = sqlite3_open (database ().c_str (), &db);
    sqlite3_busy_timeout (db, 1000);
    char * sql;
    sql = g_strdup_printf ("select text from searchresults;");
    SqliteReader reader (0);
    rc = sqlite3_exec(db, sql, reader.callback, &reader, NULL);
    g_free (sql);
    sqlite3_close (db);
    return reader.ustring0;
  }
}


void BibleTime::reloadmodules ()
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql = g_strdup_printf ("insert into reloadmodules values (1);");
  rc = sqlite3_exec(db, sql, NULL, NULL, NULL);
  g_free (sql);
  sqlite3_close (db);
}


ustring BibleTime::database ()
// Returns the filename of the database.
{
  return gw_build_filename (directories_get_temp (), "bibletime.sql");
}


void BibleTime::initialize_database ()
// Initializes the database for communication with bibledit-bibletime.
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql;
  sql = g_strdup_printf ("create table runtime (run integer);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("insert into runtime values (1);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table connected (flag integer);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table getref (book integer, chapter integer, verse string);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table setref (book integer, chapter integer, verse string);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table getmodules (flag integer);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table bibles (name string);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table commentaries (name string);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table reloadmodules (flag integer);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table searchcommand (run integer, module string, text string, selector integer);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table searchresults (text string);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sql = g_strdup_printf ("create table searchdone (flag integer);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sqlite3_close (db);
  getmodules ();
}


void BibleTime::finalize_database ()
// Finalizes the database for communication with bibledit-bibletime.
// This means we set a flag that bibledit-bibletime should terminate.
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql;
  sql = g_strdup_printf ("delete from runtime;");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sqlite3_close (db);
}


void BibleTime::getmodules ()
// Sets the flags in the database that the modules should be fetched.
{
  sqlite3 *db;
  int rc;
  rc = sqlite3_open (database ().c_str (), &db);
  sqlite3_busy_timeout (db, 1000);
  char * sql;
  sql = g_strdup_printf ("insert into getmodules values (1);");
  rc = sqlite3_exec (db, sql, NULL, NULL, NULL);
  g_free (sql);
  sqlite3_close (db);
}
