/*
** Copyright (C) 2003-2006 Teus Benschop.
**  
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**  
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**  
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**  
*/


#include "libraries.h"
#include "utilities.h"
#include <glib.h>
#include "templates.h"
#include "directories.h"
#include "gwrappers.h"
#include "versification.h"
#include "mapping.h"
#include "progresswindow.h"
#include "languages.h"
#include "keyterms.h"
#include <config.h>


void check_templates (bool gui)
/*
At some point in time templates in the form of sqlite databases were distributed
with Bibledit. But it was discovered that this could give problems when a 
database was produced on the developer's computer using a newer version of
sqlite. Users still used older versions of sqlite, and the file format 
was not always compatible, so that older versions of sqlite could not read the 
templates produced with a newer version.

The solution is to distribute templates in text files, and to let Bibledit
produces databases out of them.

A signature will be created from the available textfiles. 
This is a number calculated from their modification time stamps.
Bibledit will (re)create the templates if it finds that this signature
differs from the one that has been stored before, if it was stored at all.
*/
{
  // Storage for signature.
  unsigned int signature = 0;
  
  // Add signatures from user's templates.
  ReadFiles rf_user (directories_get_templates_user (), "", ".txt");
  for (unsigned int i = 0; i < rf_user.files.size(); i++) {
    ustring filename = gw_build_filename (directories_get_templates_user (), rf_user.files[i]);
    signature += file_get_modification_time (filename);
  }
  
  // Add signatures from Bibledit's own templates.
  ReadFiles rf_bibledit (directories_get_package_data (), "", ".txt");
  for (unsigned int i = 0; i < rf_bibledit.files.size(); i++) {
    ustring filename = gw_build_filename (directories_get_package_data (), rf_bibledit.files[i]);
    signature += file_get_modification_time (filename);
  }

  // Compare calculated and stored signature, 
  // and decide whether to (re)create the templates.
  bool createtemplates = false;
  ustring signature_filename = gw_build_filename (directories_get_templates_user (), "signature");
  ReadText rt (signature_filename, true, true);
  if (rt.lines.empty()) {
    createtemplates = true;
  } else {
    if (signature != convert_to_int (rt.lines[0]))
      createtemplates = true;
  }

  // Bail out if we don't create templates.
  if (!createtemplates) 
    return;

  // Store signature.
  rt.lines.clear();
  rt.lines.push_back (convert_to_string (signature));
  write_lines (signature_filename, rt.lines);

  // Wait a little to allow the splash screen to display.
  g_usleep (1000000);
  
  // GUI
  ProgressWindow * progresswindow = NULL;
  if (gui) progresswindow = new ProgressWindow ("Creating templates", false);

  // Clear the directory of possible incompatible old stuff.
  ustring command = "rm '" + directories_get_templates () + "'/*";
  system (command.c_str());
  
  // Create the versification templates.
  if (gui) progresswindow->set_text ("Versifications");
  vector <ustring> paths = versification_get_raw_templates_available ();
  if (gui) progresswindow->set_iterate (0, 1, paths.size());
  for (unsigned int i = 0; i < paths.size(); i++) {
    if (gui) progresswindow->iterate ();
    versification_import_textfile (versification_get_name (paths[i]), paths[i]);
  }

  // Create the mapping templates.
  if (gui) progresswindow->set_text ("Verse mappings");
  paths = mapping_get_raw_templates_available ();
  if (gui) progresswindow->set_iterate (0, 1, paths.size());
  for (unsigned int i = 0; i < paths.size(); i++) {
    if (gui) progresswindow->iterate ();
    mapping_import_textfile (mapping_get_name (paths[i]), paths[i]);
  }

  // Create the language templates.
  if (gui) progresswindow->set_text ("Languages");
  paths = language_get_raw_templates_available ();
  if (gui) progresswindow->set_iterate (0, 1, paths.size());
  for (unsigned int i = 0; i < paths.size(); i++) {
    if (gui) progresswindow->iterate ();
    language_import_textfile (language_get_name (paths[i]), paths[i]);
  }

  // Check that the "English" template is always there.
  vector<ustring> lang1 = language_get_ones_available ();
  set<ustring> lang2 (lang1.begin(), lang1.end());
  if (lang2.find ("English") == lang2.end()) {
    for (unsigned int i = 0; i < paths.size(); i++) {
      if (paths[i].find ("english") != string::npos) {
        language_import_textfile ("English", paths[i]);
        cout << "Adding English language" << endl;        
        break;
      }
    }
  }
 
  // Create the keyterms database.
  if (gui) progresswindow->set_text ("Keyterms");
  keyterms_create_database ();
  paths = keyterms_get_raw_templates_available ();
  if (gui) progresswindow->set_iterate (0, 1, paths.size() + 3);
  for (unsigned int i = 0; i < paths.size(); i++) {
    if (gui) progresswindow->iterate ();
    keyterms_import_textfile (paths[i]);
  }
  if (gui) progresswindow->iterate ();
  keyterms_import_otkey_db ();  
  if (gui) progresswindow->iterate ();
  keyterms_import_ktref_db ();  
  if (gui) progresswindow->iterate ();
  keyterms_import_ktbh_txt ();  
  
  // Destroy gui.
  if (progresswindow) delete progresswindow;
}
